import logging
from enum import StrEnum, unique
from typing import Annotated, cast

from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator

from pipeline_flow.core.models.phases import (
    ExtractPhase,
    LoadPhase,
    Phase,
    PipelinePhase,
    TransformLoadPhase,
    TransformPhase,
)


@unique
class PipelineType(StrEnum):
    """A config class that contains constants and utilities related to pipelines."""

    ETL = "ETL"
    ELT = "ELT"
    ETLT = "ETLT"


MANDATORY_PHASES_BY_PIPELINE_TYPE = {
    PipelineType.ETL: {
        PipelinePhase.EXTRACT_PHASE: True,
        PipelinePhase.TRANSFORM_PHASE: False,
        PipelinePhase.LOAD_PHASE: True,
    },
    PipelineType.ELT: {
        PipelinePhase.EXTRACT_PHASE: True,
        PipelinePhase.LOAD_PHASE: True,
        PipelinePhase.TRANSFORM_AT_LOAD_PHASE: True,
    },
    PipelineType.ETLT: {
        PipelinePhase.EXTRACT_PHASE: True,
        PipelinePhase.TRANSFORM_PHASE: False,
        PipelinePhase.LOAD_PHASE: True,
        PipelinePhase.TRANSFORM_AT_LOAD_PHASE: True,
    },
}


class Pipeline(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: Annotated[str, "Name of the pipeline job"]
    type: PipelineType
    phases: dict[PipelinePhase, Phase]

    # Optional
    description: str | None = None
    needs: str | list[str] | None = None

    # Private
    _is_executed: bool = False

    @property
    def is_executed(self) -> bool:
        return self._is_executed

    @is_executed.setter
    def is_executed(self, value: bool) -> None:
        self._is_executed = value

    @property
    def extract(self) -> ExtractPhase:
        return cast(ExtractPhase, self.phases[PipelinePhase.EXTRACT_PHASE])

    @property
    def transform(self) -> TransformPhase:
        return cast(TransformPhase, self.phases[PipelinePhase.TRANSFORM_PHASE])

    @property
    def load(self) -> LoadPhase:
        return cast(LoadPhase, self.phases[PipelinePhase.LOAD_PHASE])

    @property
    def load_transform(self) -> TransformLoadPhase:
        return cast(TransformLoadPhase, self.phases[PipelinePhase.TRANSFORM_AT_LOAD_PHASE])

    @field_validator("phases")
    @classmethod
    def validate_phase_existence(
        cls, phases: dict[PipelinePhase, Phase], info: ValidationInfo
    ) -> dict[PipelinePhase, Phase]:
        pipeline_type = info.data["type"]

        pipeline_phases = MANDATORY_PHASES_BY_PIPELINE_TYPE[pipeline_type]
        mandatory_phases = {phase for phase, is_mandatory in pipeline_phases.items() if is_mandatory}
        optional_phases = {phase for phase, is_mandatory in pipeline_phases.items() if not is_mandatory}
        # Compute missing and extra phases
        provided_phases = set(phases.keys())

        missing_phases = set(mandatory_phases) - set(provided_phases | optional_phases)
        extra_phases = set(provided_phases) - (mandatory_phases | optional_phases)

        if missing_phases or extra_phases:
            if missing_phases:
                logging.error(
                    "Validation Error: Missing phases for pipeline type '%s': %s", pipeline_type, missing_phases
                )
            if extra_phases:
                logging.warning(
                    "Validation Warning: Extra phases provided for pipeline type '%s': %s", pipeline_type, extra_phases
                )

            error_msg = (  # noqa: UP032
                "Validation Error: The provided phases do not match the required phases for pipeline type '{}'. "
                "Missing phases: {}. Extra phases: {}."
            ).format(pipeline_type, missing_phases, extra_phases)
            raise ValueError(error_msg)

        msg = f"Phase validation successful for pipeline type '{pipeline_type}'"
        logging.info(msg)
        return phases



# Standard Imports
import logging
import logging.config
import os
from pathlib import Path

# Third-party imports


# Project Imports


def setup_logger() -> None:
    log_level = os.getenv("LOG_LEVEL", "info").lower()

    config_files = {
        "info": "logging/info_logging.conf",
        "debug": "logging/debug_logging.conf",
    }

    config_file = config_files.get(log_level)

    if not config_file:
        msg = "Invalid logging level: %s. Expected 'info' or 'debug'.", log_level
        raise ValueError(msg)

    if not Path(config_file).is_file():
        msg = "Logging configuration file not found: %s", config_file
        raise FileNotFoundError(msg)

    # Load logging configuration from the selected config file
    logging.config.fileConfig(config_file, disable_existing_loggers=False)



# Standard Imports
from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta, abstractmethod
from functools import reduce
from typing import TYPE_CHECKING, Any

from pipeline_flow.common.exceptions import (
    ExtractError,
    LoadError,
    TransformError,
    TransformLoadError,
)

# Third Party Imports
# Local Imports
from pipeline_flow.common.utils import async_time_it, sync_time_it
from pipeline_flow.core.models.pipeline import Pipeline, PipelineType

# Type Imports

if TYPE_CHECKING:
    from pipeline_flow.common.type_def import ETLData, ExtractedData, TransformedData
    from pipeline_flow.core.models.phases import (
        ExtractPhase,
        LoadPhase,
        TransformLoadPhase,
        TransformPhase,
    )
    from pipeline_flow.plugins import IPlugin


def plugin_sync_executor(plugin: IPlugin, *pipeline_args: Any, **pipeline_kwargs: Any) -> ETLData:  # noqa: ANN401
    logging.debug("Executing plugin `%s`", plugin.id)
    result = plugin(*pipeline_args, **pipeline_kwargs)
    logging.debug("Finished executing plugin `%s`", plugin.id)
    return result


async def plugin_async_executor(plugin: IPlugin, *pipeline_args: Any, **pipeline_kwargs: Any) -> ETLData:  # noqa: ANN401
    logging.debug("Executing plugin `%s`", plugin.id)
    result = await plugin(*pipeline_args, **pipeline_kwargs)
    logging.debug("Finished executing plugin `%s`", plugin.id)
    return result


async def task_group_executor(
    plugins: list[IPlugin],
    *pipeline_args: Any,  # noqa: ANN401
    **pipeline_kwargs: Any,  # noqa: ANN401
) -> dict[str, ETLData]:
    async with asyncio.TaskGroup() as group:
        tasks = {
            plugin.id: group.create_task(plugin_async_executor(plugin, *pipeline_args, **pipeline_kwargs))
            for plugin in plugins
        }

    return {plugin_id: task.result() for plugin_id, task in tasks.items()}


@async_time_it
async def run_extractor(extracts: ExtractPhase) -> ExtractedData:
    results = {}

    try:
        if extracts.pre:
            await task_group_executor(extracts.pre)

        results = await task_group_executor(extracts.steps)

        # Return the single result directly.
        if len(extracts.steps) == 1:
            return results[extracts.steps[0].id]

        # Merge data if multiple steps exist.
        return plugin_sync_executor(extracts.merge, extracted_data=results)  # type: ignore[reportArgumentType] - merge will always be populated when more than 2 extracts are provided

    except Exception as e:
        error_message = "Extraction Phase Error"
        raise ExtractError(error_message, e) from e


@sync_time_it
def run_transformer(data: ExtractedData, transformations: TransformPhase) -> TransformedData:
    if not transformations.steps:
        logging.info("No transformations to run")
        return data

    try:
        transformed_data = reduce(lambda data, plugin: plugin_sync_executor(plugin, data), transformations.steps, data)
    except Exception as e:
        msg = "Transformation Phase Error"
        raise TransformError(msg, e) from e

    return transformed_data


@async_time_it
async def run_loader(data: ExtractedData | TransformedData, destinations: LoadPhase) -> None:
    if destinations.pre:
        await task_group_executor(destinations.pre)

    try:
        await task_group_executor(destinations.steps, data=data)

        if destinations.post:
            await task_group_executor(destinations.post)
    except Exception as e:
        error_message = "Load Phase Error"
        raise LoadError(error_message, e) from e


@sync_time_it
def run_transformer_after_load(transformations: TransformLoadPhase) -> None:
    try:
        for plugin in transformations.steps:
            plugin_sync_executor(plugin)

    except Exception as e:
        error_message = "Transform Load Phase Error"
        raise TransformLoadError(error_message, e) from e


class PipelineStrategy(metaclass=ABCMeta):
    @abstractmethod
    async def execute(self, pipeline: Pipeline) -> bool:
        raise NotImplementedError("This has to be implemented by the subclasses.")


class ETLStrategy(PipelineStrategy):
    async def execute(self, pipeline: Pipeline) -> bool:
        extracted_data = await run_extractor(pipeline.extract)

        # Transform (CPU-bound work, so offload to executor)
        transformed_data = await asyncio.get_running_loop().run_in_executor(
            None, run_transformer, extracted_data, pipeline.transform
        )

        await run_loader(transformed_data, pipeline.load)

        return True


class ELTStrategy(PipelineStrategy):
    async def execute(self, pipeline: Pipeline) -> bool:
        extracted_data = await run_extractor(pipeline.extract)

        await run_loader(extracted_data, pipeline.load)

        run_transformer_after_load(pipeline.load_transform)

        return True


class ETLTStrategy(PipelineStrategy):
    async def execute(self, pipeline: Pipeline) -> bool:
        extracted_data = await run_extractor(pipeline.extract)

        transformed_data = await asyncio.get_running_loop().run_in_executor(
            None, run_transformer, extracted_data, pipeline.transform
        )

        await run_loader(transformed_data, pipeline.load)

        run_transformer_after_load(pipeline.load_transform)

        return True


PIPELINE_STRATEGY_MAP = {
    PipelineType.ETL: ETLStrategy,
    PipelineType.ELT: ELTStrategy,
    PipelineType.ETLT: ETLTStrategy,
}



# Standard Imports
import asyncio
import logging

# Third Party Imports
# Project Imports
from pipeline_flow.core.executor import PIPELINE_STRATEGY_MAP
from pipeline_flow.core.models.pipeline import Pipeline
from pipeline_flow.core.parsers.yaml_parser import YamlConfig


class PipelineOrchestrator:
    """Emphasizes the role of the class in executing the pipelines."""

    def __init__(self, config: YamlConfig) -> None:
        self.concurrency = config.concurrency
        self.pipeline_queue = asyncio.Queue()
        self.semaphore = asyncio.Semaphore(config.concurrency)

    @staticmethod
    def _can_execute(pipeline: Pipeline, executed_pipelines: set[str]) -> bool:
        """A function that checks if a given pipeline is executable or has an external dependency."""
        if pipeline.needs is None:
            return True
        if isinstance(pipeline.needs, str):
            return pipeline.needs in executed_pipelines
        if isinstance(pipeline.needs, list):
            return all(need in executed_pipelines for need in pipeline.needs)
        return False

    async def pipeline_queue_producer(self, pipelines: list[Pipeline]) -> None:
        for pipeline in pipelines:
            logging.debug("Adding %s to central pipeline queue", pipeline.name)
            await self.pipeline_queue.put(pipeline)
            logging.debug("Added %s to central pipeline queue", pipeline.name)

    async def _execute_pipeline(self) -> None:
        if self.pipeline_queue.empty():
            return

        async with self.semaphore:
            while not self.pipeline_queue.empty():
                pipeline = await self.pipeline_queue.get()

                logging.info("Executing: %s ", pipeline.name)
                strategy = PIPELINE_STRATEGY_MAP[pipeline.type]
                pipeline.is_executed = await strategy().execute(pipeline)
                logging.info("Completed: %s", pipeline.name)

                self.pipeline_queue.task_done()

    async def execute_pipelines(self, pipelines: list[Pipeline]) -> set[str]:
        """Asynchronously executes parsed jobs."""
        if not pipelines:
            raise ValueError("The Pipeline list is empty. There is nothing to execute.")

        executed_pipelines = set()
        while pipelines:
            executable_pipelines = [
                pipeline for pipeline in pipelines if self._can_execute(pipeline, executed_pipelines)
            ]

            if not executable_pipelines:
                raise ValueError("Circular dependency detected!")

            # Produces a central queue of executable pipelines
            await self.pipeline_queue_producer(executable_pipelines)

            async with asyncio.TaskGroup() as tg:
                tasks = [tg.create_task(self._execute_pipeline()) for _ in range(self.concurrency)]  # noqa: F841

            executed_pipelines.update(pipeline.name for pipeline in executable_pipelines)
            pipelines = [pipeline for pipeline in pipelines if not pipeline.is_executed]

        return executed_pipelines



from __future__ import annotations

from typing import Any

from pipeline_flow.core.models.phases import (
    ExtractPhase,
    LoadPhase,
    Phase,
    PipelinePhase,
    TransformLoadPhase,
    TransformPhase,
)
from pipeline_flow.core.models.pipeline import Pipeline

PHASE_CLASS_MAP = {
    PipelinePhase.EXTRACT_PHASE: ExtractPhase,
    PipelinePhase.TRANSFORM_PHASE: TransformPhase,
    PipelinePhase.LOAD_PHASE: LoadPhase,
    PipelinePhase.TRANSFORM_AT_LOAD_PHASE: TransformLoadPhase,
}


def parse_pipelines(pipelines_data: dict[str, dict[str, Any]]) -> list[Pipeline]:
    if not pipelines_data:
        raise ValueError("No Pipelines detected.")

    return [_create_pipeline(pipeline_name, pipeline_data) for pipeline_name, pipeline_data in pipelines_data.items()]


def _create_pipeline(pipeline_name: str, phase_data: dict[str, Any]) -> Pipeline:
    """Parse a single pipeline's data and return a pipeline instance."""
    if not phase_data:
        raise ValueError("Pipeline attributes are empty")

    phases = {}

    if "phases" not in phase_data:
        raise ValueError("The argument `phases` in pipelines must be specified.")
    for phase_name, phase_details in phase_data["phases"].items():
        phases[phase_name] = _parse_phase(phase_name, phase_details)

    phase_data["phases"] = phases
    return Pipeline(name=pipeline_name, **phase_data)


def _parse_phase(phase_name: str, phase_data: dict[str, Any]) -> Phase:
    phase_pipeline: PipelinePhase = PipelinePhase(phase_name)
    phase_class = PHASE_CLASS_MAP[phase_pipeline]

    return phase_class(**phase_data)



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from pipeline_flow.common.type_def import PluginRegistryJSON


class PluginParser:
    def __init__(self: Self, plugins_payload: PluginRegistryJSON) -> None:
        self.plugins_payload = plugins_payload

    @staticmethod
    def get_all_files(paths: list[str]) -> set[str]:
        files = set()
        for path in paths:
            if os.path.isdir(path):  # noqa: PTH112
                for filename in os.listdir(path):
                    if filename.endswith(".py"):
                        full_path = Path(path) / filename
                        files.add(str(full_path))
            elif path.endswith(".py"):
                files.add(path)

        return files

    def fetch_custom_plugin_files(self: Self) -> set[str]:
        custom_payload = self.plugins_payload.get("custom", {})
        if not custom_payload:
            logging.debug("No custom plugins found in the YAML.")
            return set()

        # Gather files from dirs and individual files
        files_from_dir = self.get_all_files(custom_payload.get("dirs", set()))
        files = self.get_all_files(custom_payload.get("files", set()))

        # Combine both sets of files
        return files_from_dir.union(files)

    def fetch_community_plugin_modules(self: Self) -> set[str]:
        comm_payload = self.plugins_payload.get("community", {})
        if not comm_payload:
            logging.debug("No community plugins found in the YAML.")
            return set()

        base_module = "community.plugins."

        return {base_module + plugin for plugin in comm_payload}



# Standard Imports
from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, ClassVar

# Third Party Imports
# Project Imports
from pipeline_flow.common.utils import SingletonMeta

if TYPE_CHECKING:
    from pipeline_flow.common.type_def import PluginName, PluginPayload
    from pipeline_flow.plugins import IPlugin


class PluginRegistry(metaclass=SingletonMeta):
    """Plugin registry class for storing and retrieving plugins."""

    _registry: ClassVar[dict[PluginName, IPlugin]] = {}

    @classmethod
    def register(cls: PluginRegistry, plugin_name: PluginName, plugin_callable: IPlugin) -> None:
        """Registers a plugin in the registry.

        If the plugin is already registered, a ValueError is raised.
        """
        logging.debug("Registering plugin `%s`.", plugin_name)
        # Check if the plugin has been registered.
        if plugin_name in cls._registry:
            logging.warning("Plugin for `%s` already exists in PluginRegistry class.", plugin_name)

        cls._registry[plugin_name] = plugin_callable
        logging.debug("Plugin `%s` have been successfully registered. ", plugin_name)

    @classmethod
    def get(cls: PluginRegistry, plugin_name: PluginName) -> IPlugin:
        """Retrieve a plugin from the registry."""
        logging.debug("Retrieving plugin class for `%s`.", plugin_name)
        plugin_factory = cls._registry.get(plugin_name, None)

        if not plugin_factory:
            msg = f"Plugin class was not found for following plugin `{plugin_name}`."
            raise ValueError(msg)

        logging.debug("Plugin class '%s' has been successfully retrieved.", plugin_factory)
        return plugin_factory

    @classmethod
    def instantiate_plugin(cls: PluginRegistry, plugin_data: PluginPayload) -> IPlugin:
        """Resolve and return a single plugin instance."""
        plugin_name = plugin_data.pop("plugin", None)
        if not plugin_name:
            raise ValueError("The attribute 'plugin' is empty.")

        plugin_factory: IPlugin = cls.get(plugin_name)

        plugin_id = plugin_data.pop("id", None) or f"{plugin_name}_{uuid.uuid4().hex[:16]}"
        plugin_params = plugin_data.get("args", {})

        return plugin_factory(plugin_id=plugin_id, **plugin_params)


# Registering plugins after PluginRegistry class definition. This is done to avoid circular imports as
# the plugins are registed when they are imported from IPlugin interface.
from pipeline_flow.plugins import *  # noqa: E402, F403



from .base import (
    IExtractPlugin,
    ILoadPlugin,
    IMergeExtractPlugin,
    IPaginationHandler,
    IPlugin,
    IPreProcessPlugin,
    ISecretManager,
    ITransformLoadPlugin,
    ITransformPlugin,
)

__all__ = [
    "IExtractPlugin",
    "ILoadPlugin",
    "ILoadPlugin",
    "IMergeExtractPlugin",
    "IPaginationHandler",
    "IPlugin",
    "IPreProcessPlugin",
    "ISecretManager",
    "ITransformLoadPlugin",
    "ITransformPlugin",
]



# Standard Imports
from __future__ import annotations

import os
import re
from dataclasses import dataclass
from enum import StrEnum
from typing import TYPE_CHECKING

# Third Party Imports
import yamlcore

# Local Imports
from pipeline_flow.common.utils import SingletonMeta
from pipeline_flow.core.parsers import secret_parser

# Type Imports
if TYPE_CHECKING:
    from typing import Generator, Match, Self

    from yaml.nodes import Node

    from pipeline_flow.common.type_def import PluginRegistryJSON, StreamType


type JSON_DATA = dict

DEFAULT_CONCURRENCY = 2

ENV_VAR_YAML_TAG = "!env_var"
ENV_VAR_PATTERN = re.compile(r"\${{\s*env\.([^}]+?)\s*}}")

SECRET_YAML_TAG = "!secret"  # noqa: S105 - False Positive
SECRET_PATTERN = re.compile(r"\${{\s*secrets\.([^}]+?)\s*}}")

VARIABLE_YAML_TAG = "!variable"
VARIABLE_PATTERN = re.compile(r"\${{\s*variables\.([^}]+?)\s*}}")


class YamlAttribute(StrEnum):
    SECRETS = "secrets"
    VARIABLES = "variables"
    PIPELINES = "pipelines"
    PLUGINS = "plugins"
    CONCURRENCY = "concurrency"


@dataclass(frozen=True)
class YamlConfig(metaclass=SingletonMeta):
    concurrency: int = DEFAULT_CONCURRENCY


class ExtendedCoreLoader(yamlcore.CCoreLoader):
    """An extension of YAML 1.2 Compliant Loader to handle boolean values like `on` or `off`."""

    def __init__(self: Self, stream: StreamType) -> None:
        super().__init__(stream)
        self.secrets = {}
        self.variables = {}

    def update_variables(self: Self, new_variables: dict[str, str]) -> None:
        """Update the variables dynamically after initialization."""
        self.variables.update(new_variables)

    def update_secrets(self: Self, new_secrets: dict[str, str]) -> None:
        """Update the secrets dynamically after initialization."""
        self.secrets.update(new_secrets)

    def substitute_env_var_placeholder(self: Self, node: Node) -> str:
        """Parses a YAML node for an env var reference and replaces it with the value.

        Args:
            node (Node): A YAML node containing the `!env_var` reference added by implicit resolver.

        Raises:
            ValueError: If the environment variable is not set.

        Returns:
            str: A String with the environment variable's value.
        """
        value = node.value

        for match in ENV_VAR_PATTERN.finditer(value):
            match_group = match.group()
            env_key = match.group(1)

            env_var_value = os.environ.get(env_key, None)
            if not env_var_value:
                error_msg = f"Environment variable `{env_key}` is not set."
                raise ValueError(error_msg)

            value = value.replace(match_group, env_var_value)
        return value

    def substitute_variable_placeholder(self: Self, node: Node) -> str:
        """Parses a YAML node for a variable reference and replaces it with the value.

        If the value is a single variable, it returns the original type (int, dict, etc.).
        Else it converts the value to a string for inline substitution.

        Args:
            node (Node): A YAML node containing the `!variable` reference added by implicit resolver.

        Returns:
            str: A String with the variable's value.
        """
        value = node.value

        # Check if the entire value is just a single variable (preserve original type)
        full_match = VARIABLE_PATTERN.fullmatch(value)
        if full_match:
            variable_key = full_match.group(1)
            if variable_key not in self.variables:
                msg = f"Variable `{variable_key}` is not set."
                raise ValueError(msg)
            return self.variables[variable_key]  # Return the original type (int, dict, etc.)

        # Use re.sub with a callback for single-pass substitution of variables in strings
        def replace_match(match: Match) -> str:
            variable_key = match.group(1)
            if variable_key not in self.variables:
                msg = f"Variable `{variable_key}` is not set."
                raise ValueError(msg)
            return str(self.variables[variable_key])  # Convert to string for inline substitution

        return VARIABLE_PATTERN.sub(replace_match, value)

    def substitute_secret_placeholder(self: Self, node: Node) -> str:
        """Parses a YAML node for a secret reference and replaces it with the value.

        Args:
            node (Node): A YAML node containing the `!secret` reference added by implicit resolver.

        Returns:
            str: A String with the secret's value.
        """
        value = node.value
        match = SECRET_PATTERN.match(value)

        secret_key = match.group(1)

        if secret_key not in self.secrets:
            error_msg = (
                f"Secret `{secret_key}` is not set. "
                "Ensure that variables/secrets are defined in the first document YAML."
            )
            raise ValueError(error_msg)
        return self.secrets[secret_key]


# Register the implicit resolver to detect '${{ env.KEY }}'
ExtendedCoreLoader.add_implicit_resolver(ENV_VAR_YAML_TAG, ENV_VAR_PATTERN, None)
ExtendedCoreLoader.add_constructor(ENV_VAR_YAML_TAG, ExtendedCoreLoader.substitute_env_var_placeholder)

# Register the implicit resolver to detect '${{ secrets.KEY }}'
ExtendedCoreLoader.add_implicit_resolver(SECRET_YAML_TAG, SECRET_PATTERN, None)
ExtendedCoreLoader.add_constructor(SECRET_YAML_TAG, ExtendedCoreLoader.substitute_secret_placeholder)

# Register the implicit resolver to detect '${{ variables.KEY }}'
ExtendedCoreLoader.add_implicit_resolver(VARIABLE_YAML_TAG, VARIABLE_PATTERN, None)
ExtendedCoreLoader.add_constructor(VARIABLE_YAML_TAG, ExtendedCoreLoader.substitute_variable_placeholder)


class YamlParser:
    """YamlParser class that parses YAML content and returns the parsed data.

    Internally, it uses PyYAML Reader class to parse the YAML. That reader accepts:
        - a `bytes` object
        - a `str` object
        - a file-like object with its `read` method returning `str`
        - a file-like object with its `read` method returning `unicode`
    """

    def __init__(self: Self, stream: StreamType) -> None:
        self._stream = stream

        self._parsed_yaml = self.parse_yaml_with_context(stream)

    @property
    def yaml_body(self) -> JSON_DATA:
        """Returns the parsed YAML content as a dictionary with the root document."""
        return self._parsed_yaml

    @property
    def pipelines(self) -> JSON_DATA:
        """Returns the parsed YAML content as a dictionary with pipelines."""
        return self._parsed_yaml.get(YamlAttribute.PIPELINES, {})

    @property
    def plugins(self) -> PluginRegistryJSON | None:
        """Returns the parsed YAML content as a dictionary with plugins."""
        return self._parsed_yaml.get(YamlAttribute.PLUGINS, {})

    def stream_yaml_documents(self, loader: ExtendedCoreLoader) -> Generator[JSON_DATA, None, None]:
        """Loads all YAML documents from the stream and returns a generator of parsed data.

        Args:
            loader (ExtendedCoreLoader): The loader to use for parsing the YAML.

        Yields:
            JSON_DATA: A dictionary containing the parsed YAML content.
        """
        try:
            while loader.check_data():
                yield loader.get_data()
        finally:
            loader.dispose()  # Dispose the loader and its state

            is_file = bool(hasattr(self._stream, "read") and hasattr(self._stream, "close"))
            if is_file and not self._stream.closed:
                # Close the file if it was opened
                self._stream.close()

    def parse_yaml_with_context(self, stream: StreamType) -> JSON_DATA:
        """Loads the YAML content from the stream and returns the parsed data.

        It is a wrapper over yaml.load_all() to handle the secrets and env variables.

        Args:
            stream (StreamType): The stream to read the YAML content from.

        Returns:
            JSON_DATA: A dictionary containing the parsed YAML content.
        """
        loader = ExtendedCoreLoader(stream)

        for yaml_doc in self.stream_yaml_documents(loader):
            if YamlAttribute.SECRETS in yaml_doc:
                secrets = secret_parser(yaml_doc[YamlAttribute.SECRETS])
                loader.update_secrets(secrets)

            if YamlAttribute.VARIABLES in yaml_doc:
                variables = yaml_doc[YamlAttribute.VARIABLES]
                loader.update_variables(variables)

            if any(key in yaml_doc for key in [YamlAttribute.SECRETS, YamlAttribute.VARIABLES]):
                # Skip the document if it contains secrets or variables and move to the next document
                # to ensure that the secrets and variables are updated before parsing the next document.
                continue

            # Return the first document that doesn't contain secrets or variables.
            return yaml_doc

        # Return empty dictionary if no document is found
        return {}

    def initialize_yaml_config(self: Self) -> YamlConfig:
        """Initialize the YAML Configuration object with the default values or passed in."""
        # Create the map of attributes with their values
        attrs_map = {
            YamlAttribute.CONCURRENCY: self._parsed_yaml.get(YamlAttribute.CONCURRENCY, DEFAULT_CONCURRENCY),
        }

        # Filter out the None values
        attrs = {key: value for key, value in attrs_map.items() if value is not None}

        return YamlConfig(**attrs)



# Standard Imports
import asyncio
import logging
from http import HTTPStatus
from typing import Any, Self

# Third Party Imports
import httpx
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_random

# Local Imports
from pipeline_flow.common.type_def import PluginPayload
from pipeline_flow.core.registry import PluginRegistry
from pipeline_flow.plugins import IExtractPlugin, IPaginationHandler

JSON_DATA = dict[str, Any]


async def async_sleep(seconds: float) -> None:
    logging.debug("Retrying in %s seconds", seconds)
    await asyncio.sleep(seconds)


class RestApiAsyncExtractor(IExtractPlugin, plugin_name="rest_api_extractor"):
    """Fetches data asychronously from an API endpoint using the HTTP GET method.

    Args:
        plugin_id (str): The unique identifier of the plugin callabe. Often used for logging.
        base_url (str): The base URL of the API e.g. https://api.example.com/v1
        endpoint (str): The endpoint to fetch data from e.g. /users
        headers (dict[str, str]): A dictionary that contains headers e.g. auth token.
        pagination (PluginPayload, optional): A dict that contains plugins and args for paginations.
                                              Defaults to "page_based_pagination" plugin.
    """

    def __init__(
        self: Self,
        plugin_id: str,
        base_url: str,
        endpoint: str,
        headers: dict[str, str],
        pagination: PluginPayload | None = None,
    ) -> None:
        super().__init__(plugin_id)
        self.base_url = base_url
        self.endpoint = endpoint
        self.headers = headers

        # Fetches the pagination plugin from the registry
        # if no pagination plugin is provided, the default is "page_based".
        pagination_payload = pagination or {
            "id": f"{plugin_id}_pagination",
            "plugin": "page_based_pagination",
        }

        self.pagination_handler: IPaginationHandler = PluginRegistry.instantiate_plugin(pagination_payload)

        if not headers:
            raise ValueError("Headers must be provided for the API request.")

    @staticmethod
    def _extract_data(response_data: dict | list) -> list[JSON_DATA]:
        """Extracts data from the response JSON.

        This method handles the different ways data can be returned from an API.

        Args:
            response_data (dict | list):  The response JSON data.

        Returns:
            list[JSON_DATA]: Parsed data from the response JSON.
        """
        if isinstance(response_data, dict):
            # Standard REST API
            if "data" in response_data:
                return response_data["data"]
            return [response_data]

        if isinstance(response_data, list):
            # Direct list responses
            return response_data
        return []

    @retry(
        sleep=async_sleep,
        stop=stop_after_attempt(3),
        wait=wait_random(min=1, max=4),
        retry=retry_if_exception_type(httpx.HTTPStatusError),
        reraise=True,
    )
    async def __call__(self) -> list[JSON_DATA]:
        """Fetches data from the API endpoin asynchronously.

        Returns:
            list[JSON_DATA]: The extracted data from the API.
        """
        # TODO: Add supports for multiple endpoints with paginations async.
        results = []
        next_page_url = f"{self.base_url}/{self.endpoint}"

        # Include API key in request headers
        default_headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        default_headers.update(self.headers)

        async with httpx.AsyncClient() as client:
            while next_page_url:
                response = await client.get(url=next_page_url, headers=default_headers)

                if response.status_code != HTTPStatus.OK:
                    logging.error("Failed to retrieve data. Status code: %s", response.status_code)
                    response.raise_for_status()

                response_json = response.json()
                results.extend(self._extract_data(response_json))

                # Handle Pagination
                next_page_url = self.pagination_handler(response_json) if isinstance(response_json, dict) else None

            return results



# Standard Imports
from __future__ import annotations

import asyncio
import logging
import threading
import time
from functools import wraps
from typing import TYPE_CHECKING, Any, ClassVar

# Project Imports

# Third-party imports


# Type Imports
if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


def async_time_it[**P, R](func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
    @wraps(func)
    async def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        """Wrapper for asynchronous functions."""
        start = asyncio.get_running_loop().time()
        try:
            result = await func(*args, **kwargs)
        except Exception:
            logging.error("Error occurred while executing %s: %s")
            raise
        finally:
            total_time = asyncio.get_running_loop().time() - start
            msg = f"Time taken to execute {func.__name__} is {total_time:.4f} seconds"
            logging.info(msg)
        return result

    return inner


def sync_time_it[**P, R](func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        """Wrapper for synchronous functions."""
        start = time.time()
        try:
            result = func(*args, **kwargs)
        except Exception:
            logging.error("Error occurred while executing %s: %s")
            raise
        finally:
            total_time = time.time() - start
            msg = f"Time taken to execute {func.__name__} is {total_time:.4f} seconds"
            logging.info(msg)
        return result

    return inner


class SingletonMeta[T](type):
    _instances: ClassVar[dict] = {}

    _lock: threading.Lock = threading.Lock()

    def __call__(cls, *args: Any, **kwargs: Any) -> T:  #  noqa: ANN401
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]



# Standard Imports
from __future__ import annotations

from io import TextIOWrapper
from typing import Annotated, Any, TypedDict

# Third Party Imports
from pydantic.dataclasses import dataclass

type ExtractedData = Any
type ExtractMergedData = Any
type TransformedData = Any
type UnifiedExtractData = ExtractedData | ExtractMergedData
type ETLData = UnifiedExtractData | TransformedData


type PluginName = str
type StreamType = str | bytes | TextIOWrapper


class PluginPayload(TypedDict):
    plugin_id: str
    plugin: PluginName
    args: dict[str, Any]


class CustomPluginRegistryJSON(TypedDict):
    dirs: Annotated[list[str], "List of directories to dynamically import for custom plugins"]
    files: Annotated[list[str], "List of files to dynamically import for custom plugins"]


class PluginRegistryJSON(TypedDict):
    custom: CustomPluginRegistryJSON
    community: Annotated[list[str], "List of community plugins to import"]


@dataclass
class ExtractStageResult:
    id: str
    success: bool
    data: ExtractedData
    error: str | None = None


@dataclass
class TransformStageResult:
    id: str
    success: bool
    data: TransformedData
    error: str | None = None


@dataclass
class LoadStageResult:
    id: str
    success: bool
    error: str | None = None


@dataclass
class TransformLoadStageResult:
    id: str
    success: bool
    error: str | None = None



# Standard Imports
import logging

# # Project Imports
from pipeline_flow.common.type_def import StreamType
from pipeline_flow.common.utils import setup_logger
from pipeline_flow.core.orchestrator import PipelineOrchestrator
from pipeline_flow.core.parsers import YamlParser, parse_pipelines
from pipeline_flow.core.plugin_loader import load_plugins


async def start_orchestration(stream: StreamType) -> None:
    """Main entry point for orchestrating the pipeline flow.

    This function parses the YAML configuration, loads the plugins,
    and executes the pipelines using the PipelineOrchestrator.

    The Stream must be one of the following types:
       - `str` object: A string containing the YAML configuration.
       - 'obj' object: An object containing the YAML configuration.
       - `file` object: A file object containing the YAML configuration.

    Args:
        stream (StreamType): A stream containing the YAML configuration.
    """
    # Set up the logger configuration
    if not logging.getLogger().hasHandlers() > 0:
        setup_logger()

    # Parse YAML
    yaml_parser = YamlParser(stream)

    # Parse plugins directly within the load_plugins function
    load_plugins(yaml_parser.plugins)

    # Parse pipelines and execute them using the orchestrator
    pipelines = parse_pipelines(yaml_parser.pipelines)

    try:
        yaml_config = yaml_parser.initialize_yaml_config()
        orchestrator = PipelineOrchestrator(yaml_config)
        await orchestrator.execute_pipelines(pipelines)

    except Exception as e:
        logging.error("The following error occurred: %s", e)
        logging.error("The original cause is: %s", e.__cause__)
        raise



# Standard Imports
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

# Project Imports
from pipeline_flow.core.registry import PluginRegistry

if TYPE_CHECKING:
    from pipeline_flow.plugins import IPlugin


def serialize_plugin(value: dict) -> IPlugin:
    return PluginRegistry.instantiate_plugin(value)


def serialize_plugins(value: list) -> list[IPlugin]:
    return [PluginRegistry.instantiate_plugin(plugin_dict) for plugin_dict in value]


def unique_id_validator(steps: list[IPlugin]) -> list[IPlugin]:
    if not steps:
        logging.debug("No plugins to validate for unique ID.")
        return steps

    ids = {}

    for step in steps:
        if step.id in ids:
            raise ValueError("The `ID` is not unique. There already exists an 'ID' with this name.")

        ids[step.id] = 1

    return steps



# Standard Imports
from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ParamSpec, Self

# Third Party Imports
# Local Imports
from pipeline_flow.core.registry import PluginRegistry

if TYPE_CHECKING:
    from collections.abc import Callable

    from pipeline_flow.common.type_def import ExtractedData, ExtractMergedData, TransformedData, UnifiedExtractData

_P = ParamSpec("_P")

_PluginReturn = Any


class AsyncAdapterMixin:
    """Mixin class to run sychronous code in a new event loop thread.

    This mixin allows running synchronous code in a separate thread using
    asyncio.to_thread, so it can be awaited without blocking the main event loop.
    It is useful for wrapping blocking I/O operations (e.g. using `pandas`) in an
    async function.
    """

    async def async_wrap(
        self: Self, func: Callable[..., _PluginReturn], *args: _P.args, **kwargs: _P.kwargs
    ) -> _PluginReturn:
        """Wrap synchronous code in an async function."""
        return await asyncio.to_thread(func, *args, **kwargs)


class IPlugin:
    """Abstract base class for all plugins."""

    def __init_subclass__(
        cls,
        *,
        plugin_name: str | None = None,
        interface: bool = False,
    ) -> None:
        super().__init_subclass__()
        # If the class is an interface, do not register it.
        if interface:
            return

        if not plugin_name:
            raise ValueError("Plugin name must be provided for concrete classes.")

        # Register the plugin with the plugin registry.
        PluginRegistry.register(plugin_name, cls)

    def __init__(self: Self, plugin_id: str) -> None:
        self.id = plugin_id


class IExtractPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for extract plugins."""

    @abstractmethod
    async def __call__(self: Self) -> ExtractedData:
        """Asynchronously extract data."""
        raise NotImplementedError("Extract plugins must implement __call__()")


class IMergeExtractPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for merge-extract plugins."""

    @abstractmethod
    def __call__(self: Self, extracted_data: dict[str, ExtractedData]) -> ExtractMergedData:
        """Merge multiple extracted data sources into a single merged format."""
        raise NotImplementedError("Merge-extract plugins must implement __call__()")


class ITransformPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for transform plugins."""

    @abstractmethod
    def __call__(self: Self, data: UnifiedExtractData) -> TransformedData:
        """Transform the input data."""
        raise NotImplementedError("Transform plugins must implement __call__()")


class ILoadPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for load plugins."""

    @abstractmethod
    async def __call__(self: Self, data: UnifiedExtractData | TransformedData) -> None:
        """Asynchronously load data."""
        raise NotImplementedError("Load plugins must implement __call__()")


class ITransformLoadPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for transform-load plugins."""

    @abstractmethod
    def __call__(self: Self) -> None:
        """Sychronously transform data at the destination."""
        raise NotImplementedError("Transform-load plugins must implement __call__()")


class IPreProcessPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for pre-processing plugins."""

    @abstractmethod
    async def __call__(self: Self) -> None:
        """Pre-process data before main plugin execution."""
        raise NotImplementedError("Pre-process plugins must implement __call__()")


class IPostProcessPlugin(ABC, IPlugin, interface=True):
    """Abstract base class for post-processing plugins."""

    @abstractmethod
    async def __call__(self: Self) -> None:
        """Post-process data after main plugin execution."""
        raise NotImplementedError("Post-process plugins must implement __call__()")


class ISecretManager(ABC, IPlugin, interface=True):
    """A base class for providing authentication secrets."""

    @abstractmethod
    def __call__(self, secret_name: str) -> str:
        """A Plugin must implement this method to fetch the secret value by name."""
        raise NotImplementedError("Subclasses must implement this method.")


class IPaginationHandler(ABC, IPlugin, interface=True):
    """A base class for handling pagination in extract plugins."""

    @abstractmethod
    def __call__(self: Self, response: dict) -> str | None:
        """Asynchronously fetch data from the API endpoint and handle pagination."""
        raise NotImplementedError("Subclasses must implement this method.")



from .pipeline_parser import parse_pipelines
from .plugin_parser import PluginParser
from .secret_parser import secret_parser
from .yaml_parser import YamlParser

__all__ = ["PluginParser", "YamlParser", "parse_pipelines", "secret_parser"]



# Standard Imports
from __future__ import annotations

import logging

# Third Party Imports
import boto3
from botocore import exceptions
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential

# Local Imports
from pipeline_flow.plugins import ISecretManager


class AWSSecretManager(ISecretManager, plugin_name="aws_secret_manager"):
    """A class for fetching secrets from AWS Secret Manager."""

    def __init__(self, plugin_id: str, region: str) -> None:
        super().__init__(plugin_id)
        self.client = boto3.client("secretsmanager", region_name=region)

    @retry(
        retry=retry_if_exception_type(exceptions.EndpointConnectionError),
        wait=wait_exponential(multiplier=1, min=2, max=10),  # Exponential backoff (2s, 4s, 8s...)
        stop=stop_after_attempt(3),
        reraise=True,  # Raise exception if all retries fail
    )
    def __call__(self, secret_name: str) -> str:
        """Fetches the secret value by secret_name."""

        try:
            logging.info("Fetching secret %s from AWS Secret Manager.", secret_name)
            response = self.client.get_secret_value(SecretId=secret_name)
            logging.info("Secret fetched successfully.")
            return response["SecretString"]
        except exceptions.ClientError as e:
            error_code = e.response["Error"]["Code"]

            if error_code == "ResourceNotFoundException":
                msg = f"The requested secret {secret_name} was not found."
                logging.error(msg)
            elif error_code == "AccessDeniedException":
                msg = "Permission denied. Check IAM roles."
                logging.error(msg)

            raise



from __future__ import annotations

import importlib.util
import logging
import os
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pipeline_flow.common.type_def import PluginRegistryJSON

from pipeline_flow.core.parsers import PluginParser


def load_plugins(plugins_payload: PluginRegistryJSON | None) -> None:
    """Invoke all methods to load plugins."""
    logging.info("Starting to load plugins...")

    if not plugins_payload:
        logging.warning("No plugins to load.")
        return

    plugin_parser = PluginParser(plugins_payload)

    # Load custom plugins
    custom_files = plugin_parser.fetch_custom_plugin_files()
    load_custom_plugins(custom_files)

    # Load community plugins
    community_modules = plugin_parser.fetch_community_plugin_modules()
    load_community_plugins(community_modules)

    logging.info("All plugins loaded successfully.")


def _load_plugin_from_file(plugin_file: str) -> None:
    # Get the module name from the file and remove .py extension
    if plugin_file.startswith("/"):
        fq_module_name = plugin_file.replace(os.sep, ".")[1:-3]
    else:
        fq_module_name = plugin_file.replace(os.sep, ".")[:-3]

    # Check if the module is already loaded to avoid re-importing
    if fq_module_name in sys.modules:
        logging.debug("Module %s has been re-loaded.", fq_module_name)
        return

    try:
        logging.debug("Loading module %s", fq_module_name)
        spec = importlib.util.spec_from_file_location(fq_module_name, plugin_file)

        if not spec:
            raise ImportError(  # noqa: TRY301
                "The Spec based on following file location is empty: %s and %s plugin.", fq_module_name, plugin_file
            )

        plugin_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(plugin_module)  # type: ignore[reportOptionalMemberAccess]
        logging.info("Loaded plugin from %s as %s", plugin_file, fq_module_name)

    except ImportError:
        msg = f"Error importing plugin from `{fq_module_name}` module,"
        logging.error(msg)
        raise


def load_custom_plugins(custom_files: set[str]) -> None:
    if not custom_files:
        return
    logging.info("Planning to load all custom files: %s", custom_files)
    for custom_file in custom_files:
        _load_plugin_from_file(custom_file)

    logging.info("Loaded all custom plugins.")


def load_community_plugins(community_modules: set[str]) -> None:
    if not community_modules:
        return
    logging.info("Planning to load all community modules: %s", ",".join(list(community_modules)))
    for module in community_modules:
        _load_plugin_from_file(module)
    logging.info("Loaded all community plugins/")



# Standard Imports
from __future__ import annotations

from enum import StrEnum, unique
from typing import Annotated, Self

# Project Imports
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    model_validator,
)

from pipeline_flow.common.utils.validation import serialize_plugin, serialize_plugins, unique_id_validator
from pipeline_flow.plugins import (  # noqa: TC001 - False Positive. These are required to serialiaze the plugins.
    IExtractPlugin,
    ILoadPlugin,
    IMergeExtractPlugin,
    IPlugin,
    ITransformLoadPlugin,
    ITransformPlugin,
)

# Type Imports

# A callable class type representing all phase objects.
type Phase = ExtractPhase | TransformPhase | LoadPhase | TransformLoadPhase


@unique
class PipelinePhase(StrEnum):
    EXTRACT_PHASE = "extract"
    LOAD_PHASE = "load"
    TRANSFORM_PHASE = "transform"
    TRANSFORM_AT_LOAD_PHASE = "transform_at_load"


class ExtractPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[IExtractPlugin],
        Field(min_length=1),
        BeforeValidator(serialize_plugins),
        AfterValidator(unique_id_validator),
    ]
    pre: Annotated[
        list[IPlugin] | None,
        BeforeValidator(serialize_plugins),
    ] = None

    merge: Annotated[
        IMergeExtractPlugin | None,
        BeforeValidator(serialize_plugin),
    ] = None

    @model_validator(mode="after")
    def check_merge_condition(self: Self) -> Self:
        if self.merge and not len(self.steps) > 1:
            raise ValueError("Validation Error! Merge can only be used if there is more than one extract step.")

        if len(self.steps) > 1 and not self.merge:
            raise ValueError("Validation Error! Merge is required when there are more than one extract step.")

        return self


class TransformPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[ITransformPlugin],
        BeforeValidator(serialize_plugins),
    ]


class LoadPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[ILoadPlugin],
        Field(min_length=1),
        BeforeValidator(serialize_plugins),
    ]
    pre: Annotated[
        list[IPlugin] | None,
        BeforeValidator(serialize_plugins),
    ] = None

    post: Annotated[
        list[IPlugin] | None,
        BeforeValidator(serialize_plugins),
    ] = None


class TransformLoadPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[ITransformLoadPlugin],
        Field(min_length=1),
        BeforeValidator(serialize_plugins),
    ]



# Standard Imports
from __future__ import annotations

from typing import Self

# Third Party Imports
# Local Imports
from pipeline_flow.plugins import IPaginationHandler


class PageBasedPagination(IPaginationHandler, plugin_name="page_based_pagination"):
    def __call__(self: Self, response: dict) -> str | None:
        pagination = response.get("pagination")
        if not pagination:
            return None
        return pagination.get("next_page") if pagination.get("has_more") else None


class HATEOASPagination(IPaginationHandler, plugin_name="hateoas_pagination"):
    """Pagination strategy for APIs using HATEOAS-based links."""

    def __call__(self: Self, response: dict) -> str | None:
        links = response.get("_links", response.get("links", {}))
        return links.get("next", None)



# Standard Imports
from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, AsyncGenerator, Generator

if TYPE_CHECKING:
    from typing import Self

    from pandas import DataFrame

# Third Party Imports
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine

# Project Imports
from pipeline_flow.plugins import ILoadPlugin


class AsyncSQLAlchemyQueryLoader(ILoadPlugin, plugin_name="sqlalchemy_query_loader"):
    """A plugin that loads data into a database using SQLAlchemy query asynchronously.

    Args:
        db_user (str):  The username for the database.
        db_password (str): The password for the database.
        db_host (str): The host for the database.
        db_port (str): PORT number for the database.
        db_name (str): The name of the database.
        query (str): The query to execute uses SQLAlchemy text syntax.
        concurrency_limit (int, optional): A sephomore limit on asyncio task concurrency. Defaults to 5.
        batch_size (int, optional): The batch size. Defaults to 100000.
        driver (str, optional): The database driver. Ensure that you are using asychronous driver.
                                Defaults to "mysql+asyncmy".
    """

    def __init__(  # noqa: PLR0913
        self: Self,
        plugin_id: str,
        db_user: str,
        db_password: str,
        db_host: str,
        db_port: str,
        db_name: str,
        query: str,
        concurrency_limit: int = 5,
        batch_size: int = 100000,
        driver: str = "mysql+asyncmy",
    ) -> None:
        super().__init__(plugin_id)
        self.db_user = db_user
        self.db_password = db_password

        self.db_host = db_host
        self.db_port = db_port
        self.db_name = db_name

        self._query = query
        self._batch_size = batch_size
        self._driver = driver

        self._semaphore = asyncio.Semaphore(concurrency_limit)

        self._session_maker = self._build_async_sessionmaker()

    def _build_connection_string(self: Self) -> str:
        """A helper method that builds the connection string for the database.

        Returns:
            str: The connection string.
        """
        return f"{self._driver}://{self.db_user}:{self.db_password}@{self.db_host}:{self.db_port}/{self.db_name}"

    def _build_async_sessionmaker(self: Self) -> async_sessionmaker[AsyncSession]:
        """A helper method that builds an async session maker.

        In this link you can find more information about async context managers:
            https://docs.sqlalchemy.org/en/20/orm/session_basics.html#when-do-i-make-a-sessionmaker

        Returns:
            async_sessionmaker[AsyncSession]: An async session maker.
        """
        database_url = self._build_connection_string()
        engine = create_async_engine(database_url)
        return async_sessionmaker(engine)

    @asynccontextmanager
    async def get_async_session(self: Self) -> AsyncGenerator[AsyncSession]:
        """A context manager that yields an async session using the async session maker.

        Yields:
            Iterator[AsyncGenerator[AsyncSession, None]]: An async session.
        """
        async with self._session_maker() as session:
            try:
                yield session
            except Exception:
                await session.rollback()
                raise
            else:
                await session.commit()

    def chunk_dataframe(self: Self, df: DataFrame) -> Generator[list[dict]]:
        """A generator that chunks a pandas DataFrame into smaller dataframes.

        The chunk size is defined by the `_batch_size` attribute of the class.

        Args:
            df (pd.DataFrame): Extracted or transformed data.

        Yields:
            Generator[list[dict]]: A list of dictionaries containing the data.
        """
        for i in range(0, len(df), self._batch_size):
            yield df.iloc[i : i + self._batch_size].to_dict("records")

    async def execute_batch_query(self: Self, batch: list[dict]) -> None:
        """Executes a batch query. As per the SQLAlchemy documentation, new AsyncSession
        is created for each concurrent asyncio task.

        Here is the link to the documentation:
            https://docs.sqlalchemy.org/en/20/orm/session_basics.html#is-the-session-thread-safe-is-asyncsession-safe-to-share-in-concurrent-tasks

        Args:
            batch (list[dict]): A batch of data from the DataFrame.
        """
        async with self._semaphore:  # noqa: SIM117 - asyncio is exempt from this rule.
            async with self.get_async_session() as session:
                await session.execute(text(self._query), batch)

    async def __call__(self, data: DataFrame) -> None:
        """A method that loads data into a database using SQLAlchemy using query.

        Args:
            data (pd.DataFrame): Extracted or transformed data from the pipeline.
        """

        async with asyncio.TaskGroup() as tg:
            for batch in self.chunk_dataframe(data):
                tg.create_task(self.execute_batch_query(batch))



# Standard Imports
from __future__ import annotations

from typing import Self

# Third Party Imports

# Project Imports


class PipelineBaseError(Exception):
    """A base exception class for any pipeline errors."""

    def __init__(self: Self, message: str, original_exception: Exception | None = None) -> None:
        self.message = message
        self.original_exception = original_exception
        super().__init__(self.message)

    def __str__(self: Self) -> str:
        if self.original_exception:
            return f"{self.message} | Caused by: {self.original_exception}"
        return self.message


class ExtractError(PipelineBaseError): ...


class TransformError(PipelineBaseError): ...


class LoadError(PipelineBaseError): ...


class TransformLoadError(PipelineBaseError): ...



# Standard Imports
from typing import Any, NotRequired, Self, TypedDict

# Local Imports
from pipeline_flow.core.registry import PluginRegistry
from pipeline_flow.plugins import ISecretManager


class SecretDocument(TypedDict):
    """A type definition for the secret document."""

    plugin: str
    id: NotRequired[str]
    secret_name: str
    params: NotRequired[dict[str, Any]]


class SecretPlaceholder:
    """A class for delaying the resolution of secrets until they are needed."""

    def __init__(self: Self, secret_name: str, secret_provider: ISecretManager) -> None:
        self.secret_name = secret_name
        self.secret_provider = secret_provider

    def resolve(self: Self) -> str:
        """Fetches the secret value by secret_name."""
        return self.secret_provider(secret_name=self.secret_name)

    def __repr__(self: Self) -> str:
        """Prevent secrets from being printed or logged."""
        return f"<SecretPlaceholder: {self.secret_name} (hidden)>"


def secret_parser(document: SecretDocument) -> dict[str, SecretPlaceholder]:
    secrets = {}

    for secret, secret_data in document.items():
        plugin_provider = PluginRegistry.instantiate_plugin(secret_data)
        secrets[secret] = SecretPlaceholder(secret_data["secret_name"], plugin_provider)

    return secrets
